@@ -107,18 +107,44 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             use checkout_core::checkout::Checkout;
             use checkout_core::payment::Payment;
             use checkout_core::payment::PaymentProcessor;
-            use checkout_core::payment::handlers::{process_handler, step_handler, callback_handler, cancel_handler, get_handler as get_payment_handler};
             use checkout_core::customer::Contact;
             use checkout_core::customer::Address;
+            use checkout_core::money::Currency;
             use checkout_core::shipping::FulfillmentTypeSelection;
-            use checkout_core::server::IdParam;
-            use checkout_core::server::CreateParams;
-            use checkout_core::server::AddOrRemoveItemParams;
             use checkout_core::transaction::TransactionController;
             use checkout_core::context::{Context, ClientContext};
             use checkout_core::context::ClientContextProvider;
             use checkout_core::master_context::MasterContext;
-            use checkout_core::checkout::handlers::{create_handler, add_item_handler, remove_item_handler, confirm_items_handler, update_contact_handler, update_shipping_address_handler, update_fulfillment_type_handler, initiate_payment_handler, initiate_order_handler, get_handler as get_checkout_handler};
+            use checkout_core::order::{Order, LookupKey};
+            use checkout_core::checkout::handlers::{
+                create_handler,
+                add_item_handler,
+                remove_item_handler,
+                confirm_items_handler,
+                update_contact_handler,
+                update_shipping_address_handler,
+                update_fulfillment_type_handler,
+                initiate_payment_handler,
+                step_payment_handler as step_checkout_payment_handler,
+                initiate_order_handler,
+                release_handler,
+                get_handler as get_checkout_handler,
+            };
+            use checkout_core::payment::handlers::{
+                process_handler,
+                step_handler,
+                callback_handler,
+                cancel_handler as cancel_payment_handler,
+                get_handler as get_payment_handler,
+            };
+            use checkout_core::order::handlers::{
+                // get_handler as get_order_handler,
+                // step_payment_handler as step_order_payment_handler,
+                // confirm_handler
+                // complete_handler
+                // cancel_handler as cancel_order_handler,
+                reconcile_handler,
+            };
 
             pub struct Config {
                 pub google_application_credentials: String,
@@ -228,13 +254,88 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 });
             }
 
+            pub fn start_order_reconciliation(
+                internal_ctx: InternalContext,
+                context_provider: Arc<MasterContextProvider>,
+            ) {
+                task::spawn(async move {
+                    let subscription = internal_ctx.order_events_subscription.clone();
+
+                    while subscription.client().is_running() {
+                        println!("pulling order events messages");
+
+                        let pull = subscription.get_messages::<LookupKey>();
+
+                        match timeout(Duration::from_secs(30), pull).await {
+                            Ok(result) => match result {
+                                Ok(messages) => {
+                                    for (result, ack_id) in messages {
+                                        match result {
+                                            Ok(message) => {
+                                                println!("Received: {:?}", message);
+                                                match context_provider.new_context().await {
+                                                    Ok(mut ctx) => {
+                                                        if !ctx.start_transaction().await.is_ok() {
+                                                            continue;
+                                                        }
+                                                        let reconciliation_result = reconcile_handler(&mut ctx, &message).await;
+                                                        if !reconciliation_result.is_ok() {
+                                                            if !ctx.abort_transaction().await.is_ok() {
+                                                                // just log the failure to abort tx
+                                                                eprintln!("failed to commit tx");
+                                                            }
+                                                            continue;
+                                                        }
+                                                        if !ctx.commit_transaction().await.is_ok() {
+                                                            continue;
+                                                        }
+                                                        subscription.acknowledge_messages(vec![ack_id]).await;
+                                                    }
+                                                    Err(_) => {}
+                                                }
+                                            }
+                                            Err(e) => eprintln!("Failed to parse message: {}", e),
+                                        }
+                                    }
+                                }
+                                Err(e) => eprintln!("Failed to pull PubSub messages: {}", e)
+                            },
+                            Err(_) => eprintln!("no response in 30 seconds")
+                        }
+                    }
+                });
+            }
+
+            #[derive(Serialize, Deserialize, JsonSchema)]
+            pub struct IdParam {
+                id: String
+            }
+
+            #[derive(Serialize, Deserialize, JsonSchema)]
+            pub struct CorrelationIdParam {
+                correlation_id: String
+            }
+
+            #[derive(Serialize, Deserialize, JsonSchema)]
+            pub struct CreateCheckoutParams {
+                pub currency: Currency,
+                pub contact: Option<Contact>,
+                pub shipping_address: Option<Address>,
+            }
+
+            #[derive(Serialize, Deserialize, JsonSchema)]
+            pub struct AddOrRemoveItemParams {
+                pub sku: String,
+                pub quantity: u64,
+            }
+
             #[endpoint {
                 method = POST,
                 path = "/checkout",
             }]
-            pub async fn create_endpoint(
+            pub async fn create_checkout(
                 req_ctx: Arc<RequestContext<ServerContext>>,
-                body: TypedBody<CreateParams>,
+                body: TypedBody<CreateCheckoutParams>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
                 <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
                 <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
@@ -255,12 +356,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -269,7 +365,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = GET,
                 path = "/checkout/{id}",
             }]
-            pub async fn get_checkout_endpoint(
+            pub async fn get_checkout(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
@@ -292,12 +388,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -306,7 +397,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/add_item",
             }]
-            pub async fn add_item_endpoint(
+            pub async fn add_checkout_item(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<AddOrRemoveItemParams>,
@@ -331,12 +422,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -345,7 +431,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/remove_item",
             }]
-            pub async fn remove_item_endpoint(
+            pub async fn remove_checkout_item(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<AddOrRemoveItemParams>,
@@ -370,12 +456,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -384,7 +465,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/confirm_items",
             }]
-            pub async fn confirm_items_endpoint(
+            pub async fn confirm_checkout_items(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
@@ -408,12 +489,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -422,7 +498,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/update_contact",
             }]
-            pub async fn update_contact_endpoint(
+            pub async fn update_checkout_contact(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<Contact>,
@@ -447,12 +523,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -461,7 +532,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/update_shipping_address",
             }]
-            pub async fn update_shipping_address_endpoint(
+            pub async fn update_checkout_shipping_address(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<Address>,
@@ -486,12 +557,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -500,7 +566,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/update_fulfillment_type",
             }]
-            pub async fn update_fulfillment_type_endpoint(
+            pub async fn update_checkout_fulfillment_type(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<FulfillmentTypeSelection>,
@@ -525,12 +591,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -539,7 +600,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/initiate_payment",
             }]
-            pub async fn initiate_payment_endpoint(
+            pub async fn initiate_checkout_payment(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
@@ -564,12 +625,46 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
+                    }
+                }
+            }
+
+            #[endpoint {
+                method = PATCH,
+                path = "/checkout/{id}/step_payment",
+            }]
+            pub async fn step_checkout_payment(
+                req_ctx: Arc<RequestContext<ServerContext>>,
+                path: Path<IdParam>,
+                body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
+            ) -> Result<HttpResponseOk<Checkout<Payment<
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+            >>>, HttpError> {
+                let path_params = path.into_inner();
+                let args = body.into_inner();
+                let internal_ctx = req_ctx
+                    .context()
+                    .internal_context
+                    .clone();
+
+                let mut ctx = req_ctx
+                    .context()
+                    .context_provider
+                    .new_context()
+                    .await
+                    .unwrap();
+
+                ctx.start_transaction().await.unwrap();
+                match step_checkout_payment_handler(&internal_ctx, &mut ctx, &path_params.id, args).await {
+                    Ok(co) => {
+                        ctx.commit_transaction().await.unwrap();
+                        Ok(HttpResponseOk(co))
+                    }
+                    Err(err) => {
+                        ctx.abort_transaction().await.unwrap();
+                        Err(err.into())
                     }
                 }
             }
@@ -578,7 +673,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/checkout/{id}/initiate_order",
             }]
-            pub async fn initiate_order_endpoint(
+            pub async fn initiate_checkout_order(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
@@ -607,27 +702,54 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
+                    }
+                }
+            }
+
+            #[endpoint {
+                method = PATCH,
+                path = "/checkout/{id}/release",
+            }]
+            pub async fn release_checkout(
+                req_ctx: Arc<RequestContext<ServerContext>>,
+                path: Path<IdParam>,
+            ) -> Result<HttpResponseOk<Checkout<Payment<
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+            >>>, HttpError> {
+                let path_params = path.into_inner();
+                let internal_ctx = req_ctx
+                    .context()
+                    .internal_context
+                    .clone();
+
+                let mut ctx = req_ctx
+                    .context()
+                    .context_provider
+                    .new_context()
+                    .await
+                    .unwrap();
+
+                ctx.start_transaction().await.unwrap();
+                match release_handler(&internal_ctx, &mut ctx, &path_params.id).await {
+                    Ok(co) => {
+                        ctx.commit_transaction().await.unwrap();
+                        Ok(HttpResponseOk(co))
+                    }
+                    Err(err) => {
+                        ctx.abort_transaction().await.unwrap();
+                        Err(err.into())
                     }
                 }
             }
 
             // payment related endpoints
-            #[derive(Serialize, Deserialize, JsonSchema)]
-            pub struct CorrelationIdParam {
-                correlation_id: String
-            }
-
             #[endpoint {
                 method = GET,
                 path = "/payment/{id}",
             }]
-            pub async fn get_payment_endpoint(
+            pub async fn get_payment(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Payment<
@@ -650,12 +772,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -664,7 +781,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/payment/{id}/step",
             }]
-            pub async fn step_payment_endpoint(
+            pub async fn step_payment(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
@@ -694,12 +811,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -708,7 +820,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/payment_callback/{correlation_id}",
             }]
-            pub async fn callback_payment_endpoint(
+            pub async fn callback_payment(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<CorrelationIdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
@@ -738,12 +850,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -752,7 +859,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 method = PATCH,
                 path = "/payment/{id}/cancel",
             }]
-            pub async fn cancel_payment_endpoint(
+            pub async fn cancel_payment(
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Payment<
@@ -772,19 +879,14 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     .unwrap();
 
                 ctx.start_transaction().await.unwrap();
-                match cancel_handler(&internal_ctx, &mut ctx, &path_params.id).await {
+                match cancel_payment_handler(&internal_ctx, &mut ctx, &path_params.id).await {
                     Ok(co) => {
                         ctx.commit_transaction().await.unwrap();
                         Ok(HttpResponseOk(co))
                     }
                     Err(err) => {
                         ctx.abort_transaction().await.unwrap();
-                        Err(HttpError {
-                            status_code: StatusCode::BAD_REQUEST,
-                            error_code: Some(err.code.clone()),
-                            external_message: err.message.clone(),
-                            internal_message: err.message.clone(),
-                        })
+                        Err(err.into())
                     }
                 }
             }
@@ -796,20 +898,22 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
 
             fn api_description() -> checkout_server::ApiDescription<Self::Context> {
                 let mut api = checkout_server::ApiDescription::new();
-                api.register(checkout_server::create_endpoint).unwrap();
-                api.register(checkout_server::get_checkout_endpoint).unwrap();
-                api.register(checkout_server::add_item_endpoint).unwrap();
-                api.register(checkout_server::remove_item_endpoint).unwrap();
-                api.register(checkout_server::confirm_items_endpoint).unwrap();
-                api.register(checkout_server::update_contact_endpoint).unwrap();
-                api.register(checkout_server::update_shipping_address_endpoint).unwrap();
-                api.register(checkout_server::update_fulfillment_type_endpoint).unwrap();
-                api.register(checkout_server::initiate_payment_endpoint).unwrap();
-                api.register(checkout_server::get_payment_endpoint).unwrap();
-                api.register(checkout_server::step_payment_endpoint).unwrap();
-                api.register(checkout_server::callback_payment_endpoint).unwrap();
-                api.register(checkout_server::cancel_payment_endpoint).unwrap();
-                api.register(checkout_server::initiate_order_endpoint).unwrap();
+                api.register(checkout_server::create_checkout).unwrap();
+                api.register(checkout_server::get_checkout).unwrap();
+                api.register(checkout_server::add_checkout_item).unwrap();
+                api.register(checkout_server::remove_checkout_item).unwrap();
+                api.register(checkout_server::confirm_checkout_items).unwrap();
+                api.register(checkout_server::update_checkout_contact).unwrap();
+                api.register(checkout_server::update_checkout_shipping_address).unwrap();
+                api.register(checkout_server::update_checkout_fulfillment_type).unwrap();
+                api.register(checkout_server::initiate_checkout_payment).unwrap();
+                api.register(checkout_server::step_checkout_payment).unwrap();
+                api.register(checkout_server::initiate_checkout_order).unwrap();
+                api.register(checkout_server::release_checkout).unwrap();
+                api.register(checkout_server::get_payment).unwrap();
+                api.register(checkout_server::step_payment).unwrap();
+                api.register(checkout_server::callback_payment).unwrap();
+                api.register(checkout_server::cancel_payment).unwrap();
 
                 api
             }
@@ -837,6 +941,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 pubsub_client.spawn_token_renew(checkout_server::Duration::from_secs(15 * 60));
 
                 checkout_server::start_payment_processing(internal_context.clone(), context_provider.clone());
+                checkout_server::start_order_reconciliation(internal_context.clone(), context_provider.clone());
 
                 checkout_server::HttpServerStarter::new(
                     &checkout_server::ConfigDropshot {
